/***************************************************************************
                          s_partition_function.h  -  description
                             -------------------
    begin                : Wed Mar 15 2006
    copyright            : (C) 2006 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PARTITION_FUNCTION_H
#define PARTITION_FUNCTION_H

//#include "s_stacked_pair.h"
#include "s_hairpin_loop.h"
#include "s_internal_loop.h"
#include "s_multi_loop.h"
#include "s_multi_loop_sub.h"


class s_partition_function
{
    public:

//        friend class s_stacked_pair;
//        friend class s_internal_loop;
//        friend class s_multi_loop;

        s_partition_function (char *seq, int ignore_dangles=0);
        // The constructor                

        ~s_partition_function ();
        // The destructor

        double compute_partition_function ();
  
        void compute_base_pair_probabilities ();
        // Nov 9, 2006. Computes base pair probabilities
        // PRE: the partition function arrays have been filled

        void print_base_pair_probabilities ();

        double get_probability (int i, int j) { return p[index[i]+j-i]; } 
        
        void PS_dot_plot(char *wastlfile);        
        
        double getOneoverRT () { return oneoverRT; }
        
        // exhaustive, for verification
        double compute_partition_function_exhaustively ();        
        double Z;        
        double Zexhaustive;
        void verify_partition_function();
        
        void compute_logZ_gradient ();
        // this actually computes -RT*logZ_gradient, which is the expected counts
        
        void compute_logZ_gradient_exhaustively ();
        // this actually computes -RT*logZ_gradient, which is the expected counts
        
        int correct_gradient ();
        void copy_gradient (double *grad);
        int correct_gradient_nan ();
        void print_gradient();        
        // return 1 if no derivative is nan

        // arrays to compute the partition function
        double *up;    // (i,j) paired
        double *upm;    // (i,j) closes a multi-loop
        
        // arrays for exterior loops
        double *u_ip_jp;
        double *u_ip_ju;
        double *u_iu_jp;
        double *u_iu_ju;
        double *s1_jp;
        double *s1_ju;

        // arrays for multi-loop                
        double *u1_ip_jp;
        double *u1_ip_ju_jm1p;
        double *u1_ip_ju;
        double *u1_iu_jp;
        double *u1_iu_ju_jm1p;
        double *u1_iu_ju;
        double *s2_jp;
        double *s2_ju;
        double *s3_jp;
        double *s3_ju_jm1p;
        double *s3_ju;

              
        double *GlogZ;    // gradient of logZ
              
    private:

        char *csequence;
        int *sequence;             // the entire sequence for which we compute the energy
        int seqlen;                 // sequence length
        int *index;                // an array with indexes, such that we don't work with a 2D array, but with a 1D array of length (n*(n+1))/2
        int ignore_dangles;        // if 1, do not consider dangling ends at all - gradient will be 0.
        double oneoverRT;
        
                
        // arrays for base pair probabilities        
        double *p;    // base pair probabilities        
//         double *pm;    // used if ignore_dangles
//         double *pm1;    // used if ignore_dangles
//         double *pm2;    // used if ignore_dangles
        double *pmnod3_needmidd3;
        double *pmnod3_noneedmidd3;
        double *pmd3_needmidd3;
        double *pmd3_noneedmidd3;
        double *pm1nod3_needendd3;
        double *pm1d3_needendd3;
        
        // the next 4 are necessary to compute the derivative wrt multi_free_base_penalty
        double *pm2d5_needmidd5;
        double *pm2d5_noneedmidd5;
        double *pm2nod5_needmidd5;
        double *pm2nod5_noneedmidd5;        
        
        void initialize_arrays ();
        
        void compute_u_ip_jp (int i, int j);
        void compute_u_ip_ju (int i, int j);
        void compute_u_iu_jp (int i, int j);
        void compute_u_iu_ju (int i, int j);
        void compute_s1_jp (int i, int j);
        void compute_s1_ju (int i, int j);
        
        void compute_up (int i, int j);
        void compute_upm (int i, int j);
                
        void compute_u1_ip_jp (int i, int j);
        void compute_u1_ip_ju_jm1p (int i, int j);
        void compute_u1_ip_ju (int i, int j);
        void compute_u1_iu_jp (int i, int j);
        void compute_u1_iu_ju_jm1p (int i, int j);
        void compute_u1_iu_ju (int i, int j);
        
        void compute_s2_jp (int i, int j);
        void compute_s2_ju (int i, int j);
        
        void compute_s3_jp (int i, int j);
        void compute_s3_ju_jm1p (int i, int j);
        void compute_s3_ju (int i, int j);
        
        double exp_AUpenalty (int i, int j);
        double exp_dangle5 (int i, int j, int k);
        double exp_dangle3 (int i, int j, int k);        
        double fd3 (int jplus1, int h, int l);
        void compute_p (int h, int l);
//         void compute_pm (int h, int l);    // used if ignore_dangles
//         void compute_pm1 (int h, int l);    // used if ignore_dangles
//         void compute_pm2 (int h, int l);    // used if ignore_dangles
        void compute_pmnod3_needmidd3 (int h, int l);   
        void compute_pmnod3_noneedmidd3 (int h, int l);   
        void compute_pmd3_needmidd3 (int h, int l);
        void compute_pmd3_noneedmidd3 (int h, int l);
        void compute_pm1nod3_needendd3 (int h, int l);
        void compute_pm1d3_needendd3 (int h, int l);
        
        void compute_pm2d5_needmidd5 (int h, int j);
        void compute_pm2d5_noneedmidd5 (int h, int j);
        void compute_pm2nod5_needmidd5 (int h, int j);
        void compute_pm2nod5_noneedmidd5 (int h, int j);                
        
        double *pexhaustive;
        double *uexhaustive;
        double *upexhaustive;
        int has_base_pair (int i, int j, char *structure);
        int identical_structure (int i, int j, char *structure1, char *structure2);
        double exp_free_energy_partial (int i, int j, char *sequence, char *structure, int removeAU);
                
        double *GlogZexhaustive;
        //double eAU; // exp AUpenalty, so that we only compute it once. - seems the same

//         // I'm using the terminology from Ding & Lawrence NAR 2003, "A statistical sampling algorithm for RNA secondary structure prediction"
//         double *u;     // partition function on region i,j (can have anything)
//         //double *unod;  // partition function on region i,j (can have anything), but which does not consider the dangling ends at all
//         double *unod5;  // partition function on region i,j (can have anything), but which does not consider the dangling ends at the most 5' end
//         double *unod3;  // partition function on region i,j (can have anything), but which does not consider the dangling ends at the most 3' end
//         double *unod35;  // partition function on region i,j (can have anything), but which does not consider the dangling ends at the most 3' end and at the most 5' end
// 
//         double *s1;    // helper
//         //double *s1nod;    // helper corresponding to unod
//         double *s1nod5;    // helper corresponding to unod5
//         double *s1nod3;    // helper corresponding to unod3
//         double *s1nod35;    // helper corresponding to unod35
// 
//         // arrays for multi-loops
// 
//         double *s2;    // helper
//         double *s2nod3;    // helper
//         double *u1;    // partial multi-loop
//         double *u1nod5;    // partial multi-loop, no dangling ends to the 5' end
//         double *u1nod3;    // partial multi-loop, no dangling ends to the 3' end
//         double *u1nod35;    // partial multi-loop, no dangling ends to the 5' and 3' ends
//
//         double *s3;    // helper
//         double *s3nod5;    // helper, no dangling end to the 5' end
//         double *s3nod3;    // helper, no dangling end to the 3' end
//         double *s3nod35;    // helper, no dangling end to the 5' and 3' ends
// 
//         void compute_u (int i, int j);
//         //void compute_unod (int i, int j);
//         void compute_unod3 (int i, int j);
//         void compute_unod5 (int i, int j);
//         void compute_unod35 (int i, int j);
// 
//         void compute_u1 (int i, int j);
//         void compute_u1nod5 (int i, int j);
//         void compute_u1nod3 (int i, int j);
//         void compute_u1nod35 (int i, int j);
// 
//         void compute_s1 (int h, int j);
//         //void compute_s1nod (int h, int j);
//         void compute_s1nod5 (int h, int j);
//         void compute_s1nod3 (int h, int j);
//         void compute_s1nod35 (int h, int j);
//         void compute_s2 (int h, int j);        
//         void compute_s2nod3 (int h, int j);        
//         void compute_s3 (int h, int j);
//         void compute_s3nod5 (int h, int j);
//         void compute_s3nod3 (int h, int j);
//         void compute_s3nod35 (int h, int j);
        
};



#endif


